/* -*- mode: C; c-file-style: "stroustrup"; c-basic-offset: 4; -*-
 *
 * This file is part of the Ruby binding of Buddy.
 *
 * This Ruby binding is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This binding is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this binding; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA.
 *
 * Author: Alexandre David.
 * Date: 2006-04-05.
 *
 */

#include <stdio.h>
#include <ruby.h>
#include <bdd.h>
#include <fdd.h>
#include <bvec.h>
#include <assert.h>

#ifdef MINGW32
#define DLLEXPORT __declspec(dllexport)
#else
#define DLLEXPORT
#endif

// Local constants.

static VALUE cBDD;      // The class BDD.
static VALUE cPair;     // The class BDD::Pair.
static VALUE cOperator; // The class BDD::Operation.
static VALUE cBVec;     // The class BVec.
static VALUE cFDD;      // The class FDD.
static ID SIZE_ID;      // ID of string "size".
static ID SYSTEM_ID;    // ID of string "system".
static ID LABEL_ID;     // ID of string "label".
static ID START_ID;     // ID of string "start".

static const char *bdd_OP[] =
{
    "AND",      // bddop_and
    "XOR",      // bddop_xor
    "OR",       // bddop_or
    "NAND",     // bddop_nand
    "NOR",      // bddop_nor
    "IMPLY",    // bddop_imp
    "BIIMPLY",  // bddop_biimp
    "DIFF",     // bddop_diff
    "LESS",     // bddop_less
    "INVIMPLY"  // bddop_invimp
};

#define NB_OP (sizeof(bdd_OP)/sizeof(char*))

// Local flag.

static int print_false = 0;

// Provided by the C++/C wrapper.
int  visited_has(int x);
void visited_clear();
void visited_insert(int x);

// Useful macros.

#define VALUE2BDD(VALUE, NAME)       \
Data_Get_Struct((VALUE), BDD, NAME); \
assert(NAME)

#define VALUE2BVEC(VALUE, NAME)      \
Data_Get_Struct((VALUE), BVEC, NAME);\
assert(NAME)

#define DEF_FDD(VALUE, NAME)         \
union { void *dummy; int fdd; } NAME;\
Data_Get_Struct((VALUE), void, NAME.dummy);

#define DEF_BDD(VALUE, NAME) \
BDD *NAME;                   \
VALUE2BDD(VALUE, NAME)

#define DEF_BVEC(VALUE, NAME) \
BVEC *NAME;                   \
VALUE2BVEC(VALUE, NAME)

#define DEF_BDDPAIR(VALUE, NAME)       \
bddPair *NAME;                         \
Data_Get_Struct(VALUE, bddPair, NAME); \
assert(NAME)

#define BDDPAIR2VALUE(PTR) \
Data_Wrap_Struct(cPair, NULL, delete_bddpair, PTR)

#define OPERATOR2VALUE(OP) \
Data_Wrap_Struct(cOperator, NULL, NULL, (void*) OP)

#define FDD2VALUE(F) \
Data_Wrap_Struct(cFDD, NULL, NULL, (void*) F)

// Deallocation hooks.

static void delete_bdd(void *vbdd)
{
    BDD *bdd = (BDD*) vbdd;
    assert(bdd);
    bdd_delref(*bdd);
    free(bdd);
}

static void delete_bddpair(void *vbddpair)
{
    assert(vbddpair);
    bdd_freepair((bddPair*) vbddpair);
}

static void delete_bvec(void *vbvec)
{
    BVEC *bvec = (BVEC*) vbvec;
    assert(bvec);
    bvec_delref(*bvec);
    bvec_free(*bvec);
    free(bvec);
}

// Allocation functions.

static BDD* new_BDD()
{
    BDD *bdd = (BDD*) malloc(sizeof(BDD));
    if (!bdd)
    {
	rb_raise(rb_eNoMemError, "BDD not allocated!");
    }
    return bdd;
}

static bddPair* new_bddPair()
{
    bddPair *bddpair = bdd_newpair();
    if (!bddpair)
    {
	rb_raise(rb_eNoMemError, "BDDPair not allocated!");
    }
    return bddpair;
}

static BVEC* new_BVEC()
{
    BVEC *bvec = (BVEC*) malloc(sizeof(BVEC));
    if (!bvec)
    {
	rb_raise(rb_eNoMemError, "BVec not allocated!");
    }
    return bvec;
}

// Conversion (+allocation) of a BDD/bddPair to VALUE.

static VALUE bdd2value(BDD arg)
{
    BDD *bdd = new_BDD();
    bdd_addref(*bdd = arg);
    return Data_Wrap_Struct(cBDD, NULL, delete_bdd, bdd);
}

static VALUE bvec2value(BVEC arg)
{
    BVEC *bvec = new_BVEC();
    *bvec = arg;
    // Already ref counted so don't add ref.
    return Data_Wrap_Struct(cBVec, NULL, delete_bvec, bvec);
}

// Assignment with care with reference counting.

static void assign_bdd(BDD *dst, BDD arg)
{
    BDD old;
    assert(dst);

    // bdd_addref before bdd_delref in case *dst == arg.
    old = *dst;
    bdd_addref(*dst = arg);
    bdd_delref(old);
}

// Conversions + checks.

static BDD check_bdd_node(BDD bdd)
{
    if (bdd == bddfalse || bdd == bddtrue)
    {
	rb_raise(rb_eArgError,
		 "BDD %s is a terminal node.",
		 bdd == bddtrue ? "true" : "false");
    }
    return bdd;
}

static VALUE cBDD_var(VALUE self)
{
    DEF_BDD(self, bdd);
    return INT2NUM(bdd_var(check_bdd_node(*bdd)));
}

static int check_id(VALUE arg) // NUM or BDD => int
{
    int i = NUM2INT(rb_obj_is_kind_of(arg, cBDD) ? cBDD_var(arg) : arg);
    if (i < 0 || i >= bdd_varnum())
    {
	rb_raise(rb_eArgError, "Unknown variable ID (%d).", i);
    }
    return i;
}

static void check_bvecs(BVEC *l, BVEC *r)
{
    assert(l && r);
    if (l->bitnum != r->bitnum)
    {
	rb_raise(rb_eArgError, "Different number of bits (%d != %d).",
		 l->bitnum, r->bitnum);
    }
}

static int value2operator(VALUE arg)
{
    union { void *dummy; int value; } op;
    Data_Get_Struct(arg, void, op.dummy);
    if (op.value < 0 || op.value >= NB_OP)
    {
	rb_raise(rb_eTypeError, "Invalid Operator value.");
    }
    return op.value;
}

// Class methods of BDD.

static VALUE cBDD_new(int argc, VALUE *argv, VALUE classValue)
{
    BDD *bdd = new_BDD();
    VALUE instance;
    *bdd = 0; // Empty BDD.
    instance = Data_Wrap_Struct(classValue, NULL, delete_bdd, bdd);
    rb_obj_call_init(instance, argc, argv);
    return instance;
}

static VALUE cBDD_print_stat(VALUE classValue)
{
    bdd_printstat();
    printf("\n");
    return classValue;
}

static VALUE cBDD_true(VALUE classValue)
{
    return bdd2value(bddtrue);
}

static VALUE cBDD_false(VALUE classValue)
{
    return bdd2value(bddfalse);
}

static VALUE cBDD_setvarnum(VALUE classValue, VALUE arg)
{
    int nb = NUM2INT(arg);
    if (nb > bdd_varnum() && bdd_setvarnum(nb))
    {
	rb_raise(rb_eStandardError,
		 "Could not increase the number of variables to %d.",
		 nb);
    }
    return INT2NUM(bdd_varnum());
}

static VALUE cBDD_getvarnum(VALUE classValue)
{
    return INT2NUM(bdd_varnum());
}

static VALUE cBDD_get_var(VALUE classValue, VALUE arg)
{
    return bdd2value(bdd_ithvar(check_id(arg)));
}

static VALUE cBDD_get_nvar(VALUE classValue, VALUE arg)
{
    return bdd2value(bdd_nithvar(check_id(arg)));
}

static VALUE cBDD_new_var(VALUE classValue)
{
    if (bdd_extvarnum(1) < 0)
    {
	rb_raise(rb_eStandardError,
		 "Could not increase the number of variables.");
	return Qnil;
    }
    else
    {
	return bdd2value(bdd_ithvar(bdd_varnum()-1));
    }
}

static VALUE cBDD_make_set(VALUE classValue, VALUE arg)
{
    int i, size = NUM2INT(rb_funcall(arg, SIZE_ID, 0));
    int id[size];
    for(i = 0; i < size; ++i)
    {
	id[i] = check_id(rb_ary_entry(arg, i));
    }
    return bdd2value(bdd_makeset(id, size));
}

static BDD load_bdd(VALUE arg)
{
    char *filename = StringValueCStr(arg);
    BDD bdd;
    int result = bdd_fnload(filename, &bdd);
    if (result)
    {
	rb_raise(rb_eIOError, "Load(%s): %s.",
		 filename, bdd_errstring(result));
    }
    return bdd;
}

static VALUE cBDD_load(VALUE classValue, VALUE arg)
{
    return bdd2value(load_bdd(arg));
}

static BDD cBDD_swapvar(VALUE classValue, VALUE v1, VALUE v2)
{
    bdd_swapvar(check_id(v1), check_id(v2));
    return classValue;
}

static VALUE cBDD_nxor(int argc, VALUE *argv, VALUE classValue)
{
    int one_array = argc == 1 && rb_obj_is_kind_of(argv[0], rb_cArray);
    int i, size = one_array ? NUM2INT(rb_funcall(argv[0], SIZE_ID, 0)) : argc;
    bdd id[size];
    BDD *result;
    BDD disj = bddfalse;
    BDD nconj = bddtrue;

    for(i = 0; i < size; ++i)
    {
	DEF_BDD(one_array ? rb_ary_entry(argv[0], i) : argv[i], bddi);
	id[i] = *bddi;
    }

    for(i = size-1; i >= 0; --i)
    {
	BDD tmp = bdd_addref(bdd_or(bdd_and(id[i], nconj),
				    bdd_and(bdd_not(id[i]), disj)));
	bdd_delref(disj);
	disj = tmp;
	if (i > 0)
	{
	    tmp = bdd_addref(bdd_and(bdd_not(id[i]), nconj));
	    bdd_delref(nconj);
	    nconj = tmp;
	}
    }

    bdd_delref(nconj);
    result = new_BDD();
    *result = disj;
    return Data_Wrap_Struct(cBDD, NULL, delete_bdd, result);
}

static VALUE cBDD_setvarorder(VALUE classValue, VALUE arg)
{
    // Allocate IDs on the stack (should be fine) and check size.

    int i, size = NUM2INT(rb_funcall(arg, SIZE_ID, 0));
    int id[size];
    char check[size];

    if (size != bdd_varnum())
    {
	rb_raise(rb_eArgError,
		 "Size (%d) must match the number of variables (%d).",
		 size, bdd_varnum());
    }

    // Convert Ruby Array to C int[] and check for duplicate IDs.

    memset(check, 0, size);
    for(i = 0; i < size; ++i)
    {
        int j = id[i] = check_id(rb_ary_entry(arg, i));
	if (check[j])
	{
	    rb_raise(rb_eArgError, "ID %d is duplicated.", j);
	}
	else
	{
	    check[j] = 1;
	}
    }

    // Change the variable ordering.
    bdd_setvarorder(id);

    return classValue;
}

#define REORDER(METHOD)                                  \
static VALUE cBDD_reorder_##METHOD(VALUE classValue)     \
{                                                        \
    bdd_reorder(BDD_REORDER_##METHOD);                   \
    return classValue;                                   \
}
#define AUTOREORDER(METHOD)                              \
static VALUE cBDD_autoreorder_##METHOD(VALUE classValue) \
{                                                        \
    bdd_autoreorder(BDD_REORDER_##METHOD);               \
    return classValue;                                   \
}
#define DEF_REORDER(METHOD) REORDER(METHOD) AUTOREORDER(METHOD)

    DEF_REORDER(WIN2)
    DEF_REORDER(WIN2ITE)
    DEF_REORDER(WIN3)
    DEF_REORDER(WIN3ITE)
    DEF_REORDER(SIFT)
    DEF_REORDER(SIFTITE)
    DEF_REORDER(RANDOM)
    AUTOREORDER(NONE)

// Instance methods of BDD.

static VALUE cBDD_initialize(int argc, VALUE *argv, VALUE self)
{
    DEF_BDD(self, bdd);

    // Copy constructor.

    if (argc == 1 && rb_obj_is_kind_of(argv[0], cBDD))
    {
	DEF_BDD(argv[0], arg);
	assign_bdd(bdd, *arg);
    }
    else if (argc > 1)
    {
	rb_raise(rb_eArgError, "0 or 1 BDD expected!");
    }
    return self;
}

/* Better to use each/sat_one and print what the user wants.

static VALUE cBDD_print_all(VALUE self)
{
    DEF_BDD(self, bdd);
    bdd_printset(*bdd);
    printf("\n");
    return self;
}

static VALUE cBDD_print_one(VALUE self)
{
    bdd sat;
    DEF_BDD(self, bdd);
    bdd_printset(sat = bdd_satone(*bdd));
    bdd_delref(sat);
    printf("\n");
    return self;
}

*/

static VALUE cBDD_node_count(VALUE self)
{
    DEF_BDD(self, bdd);
    return INT2NUM(bdd_nodecount(*bdd));
}

static VALUE cBDD_path_count(VALUE self)
{
    DEF_BDD(self, bdd);
    return rb_float_new(bdd_pathcount(*bdd));
}

static VALUE cBDD_id(VALUE self)
{
    DEF_BDD(self, bdd);
    return INT2NUM(*bdd);
}

static VALUE cBDD_equal(VALUE self, VALUE arg)
{
    BDD *self_bdd, *arg_bdd;
    VALUE2BDD(self, self_bdd);
    VALUE2BDD(arg, arg_bdd);
    return *self_bdd == *arg_bdd ? Qtrue : Qfalse;
}

static VALUE cBDD_is_true(VALUE self)
{
    DEF_BDD(self, bdd);
    return *bdd == bddtrue ? Qtrue : Qfalse;
}

static VALUE cBDD_is_false(VALUE self)
{
    DEF_BDD(self, bdd);
    return *bdd == bddfalse ? Qtrue : Qfalse;
}

static VALUE cBDD_is_var(VALUE self)
{
    DEF_BDD(self, bdd);
    return
	*bdd != bddfalse && *bdd != bddtrue &&
	bdd_high(*bdd) == bddtrue && bdd_low(*bdd) == bddfalse
	? Qtrue : Qfalse;
}

static VALUE cBDD_is_nvar(VALUE self)
{
    DEF_BDD(self, bdd);
    return
	*bdd != bddfalse && *bdd != bddtrue &&
	bdd_high(*bdd) == bddfalse && bdd_low(*bdd) == bddtrue
	? Qtrue : Qfalse;
}

static VALUE cBDD_low(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_low(check_bdd_node(*bdd)));
}

static VALUE cBDD_high(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_high(check_bdd_node(*bdd)));
}

// Declaration for printdot_rec.
static void print_handler(FILE *out, int var);

// Printing from Iben-1.1 modified.

static void printdot_rec(FILE* ofile, BDD b)
{
    if (b == bddfalse || b == bddtrue || visited_has(b))
    {
	return;
    }

    fprintf(ofile, "%d [label=\"", b);
    print_handler(ofile, bdd_var(b));
    fprintf(ofile, "\"];\n");

    if (print_false || bdd_low(b) != bddfalse)
    {
        fprintf(ofile, "%d -> %d [style=dotted];\n", b, bdd_low(b));
    }
    if (print_false || bdd_high(b) != bddfalse)
    {
        fprintf(ofile, "%d -> %d [style=filled];\n", b, bdd_high(b));
    }
    visited_insert(b);

    printdot_rec(ofile, bdd_low(b));
    printdot_rec(ofile, bdd_high(b));
}
  
static void printdot(FILE* ofile, BDD r)
{
    fprintf(ofile, "digraph G {\n");

    if (print_false && r != bddtrue)
    {
	fprintf(ofile, "0 [shape=box, label=\"0\", style=filled, shape=box, height=0.3, width=0.3];\n");
    }
    if (r != bddfalse)
    {
	fprintf(ofile, "1 [shape=box, label=\"1\", style=filled, shape=box, height=0.3, width=0.3];\n");
    }

    printdot_rec(ofile, r);
    fprintf(ofile, "}\n");

    visited_clear();
}

static VALUE cBDD_print_false(VALUE self, VALUE arg)
{
    print_false = arg != Qfalse;
    return arg;
}

#define DOT_CMD "dot -Tps -Gcenter=1 -Gsize=\"7,10\" > "

static VALUE cBDD_to_ps(VALUE self, VALUE arg)
{
    DEF_BDD(self, bdd);
    const char *filename = StringValueCStr(arg);
    int size = sizeof(DOT_CMD)+strlen(filename);
    char cmd[size];
    FILE *stream;

    snprintf(cmd, size, DOT_CMD"%s",filename);
    stream = popen(cmd, "w");
    if (!stream)
    {
	rb_raise(rb_eStandardError, "Could not start dot.");
    }
    else
    {
	printdot(stream, *bdd);
	pclose(stream);
    }
    return self;
}

static VALUE cBDD_load_self(VALUE self, VALUE arg)
{
    DEF_BDD(self, bdd);
    assign_bdd(bdd, load_bdd(arg));
    return self;
}

static VALUE cBDD_support(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_support(*bdd));
}

static VALUE cBDD_save(VALUE self, VALUE arg)
{
    char *filename = StringValueCStr(arg);
    int result;
    DEF_BDD(self, bdd);
    result = bdd_fnsave(filename, *bdd);
    if (result)
    {
	rb_raise(rb_eIOError, "Save(%s): %s.",
		 filename, bdd_errstring(result));
    }
    return self;
}

#define DEF_FUNCTION(NAME)                                 \
static VALUE cBDD_##NAME(VALUE self, VALUE arg)            \
{                                                          \
    BDD *self_bdd, *arg_bdd;                               \
    VALUE2BDD(self, self_bdd);                             \
    VALUE2BDD(arg, arg_bdd);                               \
    return bdd2value(bdd_##NAME(*self_bdd, *arg_bdd));     \
}                                                          \
static VALUE cBDD_##NAME##_self(VALUE self, VALUE arg)     \
{                                                          \
    BDD *self_bdd, *arg_bdd;                               \
    VALUE2BDD(self, self_bdd);                             \
    VALUE2BDD(arg, arg_bdd);                               \
    assign_bdd(self_bdd, bdd_##NAME(*self_bdd, *arg_bdd)); \
    return self;                                           \
}

    DEF_FUNCTION(simplify)
    DEF_FUNCTION(restrict)
    DEF_FUNCTION(exist)
    DEF_FUNCTION(forall)
    DEF_FUNCTION(unique)
    DEF_FUNCTION(constrain)

static VALUE cBDD_replace(VALUE self, VALUE arg)
{
    BDD *bdd;
    DEF_BDDPAIR(arg, bddpair);
    VALUE2BDD(self, bdd);
    return bdd2value(bdd_replace(*bdd, bddpair));
}

static VALUE cBDD_replace_self(VALUE self, VALUE arg)
{
    BDD *bdd;
    DEF_BDDPAIR(arg, bddpair);
    VALUE2BDD(self, bdd);
    assign_bdd(bdd, bdd_replace(*bdd, bddpair));
    return self;
}

static VALUE cBDD_compose(VALUE self, VALUE arg, VALUE var)
{
    BDD *self_bdd, *arg_bdd;
    VALUE2BDD(self, self_bdd);
    VALUE2BDD(arg, arg_bdd);
    return bdd2value(bdd_compose(*self_bdd, *arg_bdd, check_id(var)));
}

static VALUE cBDD_compose_self(VALUE self, VALUE arg, VALUE var)
{
    BDD *self_bdd, *arg_bdd;
    VALUE2BDD(self, self_bdd);
    VALUE2BDD(arg, arg_bdd);
    assign_bdd(self_bdd, bdd_compose(*self_bdd, *arg_bdd, check_id(var)));
    return self;
}

static VALUE cBDD_sim_compose(VALUE self, VALUE arg)
{
    BDD *self_bdd;
    DEF_BDDPAIR(arg, arg_pair);
    VALUE2BDD(self, self_bdd);
    return bdd2value(bdd_veccompose(*self_bdd, arg_pair));
}

static VALUE cBDD_sim_compose_self(VALUE self, VALUE arg)
{
    BDD *self_bdd;
    DEF_BDDPAIR(arg, arg_pair);
    VALUE2BDD(self, self_bdd);
    assign_bdd(self_bdd, bdd_veccompose(*self_bdd, arg_pair));
    return self;
}

static VALUE cBDD_ite(VALUE f, VALUE g, VALUE h)
{
    BDD *f_bdd, *g_bdd, *h_bdd;
    VALUE2BDD(f, f_bdd);
    VALUE2BDD(g, g_bdd);
    VALUE2BDD(h, h_bdd);
    return bdd2value(bdd_ite(*f_bdd, *g_bdd, *h_bdd));
}

static VALUE cBDD_ite_self(VALUE f, VALUE g, VALUE h)
{
    BDD *f_bdd, *g_bdd, *h_bdd;
    VALUE2BDD(f, f_bdd);
    VALUE2BDD(g, g_bdd);
    VALUE2BDD(h, h_bdd);
    assign_bdd(f_bdd, bdd_ite(*f_bdd, *g_bdd, *h_bdd));
    return f;
}

static void allsat_each_handler(char *varset, int size)
{
    VALUE arg = rb_ary_new2(size);
    int i;
    for(i = 0; i < size; ++i)
    {
	rb_ary_store(arg, i, varset[i] < 0 ? Qnil : (varset[i] ? Qtrue : Qfalse));
    }
    rb_yield(arg);
}

static VALUE cBDD_each(VALUE self)
{
    DEF_BDD(self, bdd);
    bdd_allsat(*bdd, allsat_each_handler);
    return self;
}

static VALUE cBDD_satcount(VALUE self)
{
    DEF_BDD(self, bdd);
    return rb_float_new(bdd_satcount(*bdd));
}

static VALUE cBDD_satcountln(VALUE self)
{
    DEF_BDD(self, bdd);
    return rb_float_new(bdd_satcountln(*bdd));
}

static VALUE cBDD_satone(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_satone(*bdd));
}

static VALUE cBDD_fullsatone(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_fullsatone(*bdd));
}

static VALUE cBDD_not(VALUE self)
{
    DEF_BDD(self, bdd);
    return bdd2value(bdd_not(*bdd));
}

static VALUE cBDD_not_self(VALUE self)
{
    DEF_BDD(self, bdd);
    assign_bdd(bdd, bdd_not(*bdd));
    return self;
}

// Const and self-modifying binary operators.

static VALUE const_op(VALUE self, VALUE arg, int op)
{
    BDD *self_bdd, *arg_bdd;
    VALUE2BDD(self, self_bdd);
    VALUE2BDD(arg, arg_bdd);
    return bdd2value(bdd_apply(*self_bdd, *arg_bdd, op));
}

static VALUE self_op(VALUE self, VALUE arg, int op)
{
    BDD *self_bdd, *arg_bdd;
    VALUE2BDD(self, self_bdd);
    VALUE2BDD(arg, arg_bdd);
    assign_bdd(self_bdd, bdd_apply(*self_bdd, *arg_bdd, op));
    return self;
}

#define DEF_OP(OP)                                   \
static VALUE cBDD_##OP(VALUE self, VALUE arg)        \
{                                                    \
    return const_op(self, arg, bddop_##OP);          \
}                                                    \
static VALUE cBDD_##OP##_self(VALUE self, VALUE arg) \
{                                                    \
    return self_op(self, arg, bddop_##OP);           \
}

    DEF_OP(and)
    DEF_OP(xor)
    DEF_OP(or)
    DEF_OP(nand)
    DEF_OP(nor)
    DEF_OP(imp)
    DEF_OP(biimp)
    DEF_OP(diff)
    DEF_OP(less)
    DEF_OP(invimp)

static VALUE cBDD_apply(VALUE self, VALUE arg, VALUE op)
{
    return const_op(self, arg, value2operator(op));
}

static VALUE cBDD_apply_self(VALUE self, VALUE arg, VALUE op)
{
    return self_op(self, arg, value2operator(op));
}

#define DEF_APPLY_OPERATOR_QUANTIFIER(APPLY)                                     \
static VALUE cBDD_##APPLY(VALUE left, VALUE right, VALUE op, VALUE var)           \
{                                                                                  \
    BDD *left_bdd, *right_bdd, *var_bdd;                                            \
    VALUE2BDD(left, left_bdd);                                                       \
    VALUE2BDD(right, right_bdd);                                                      \
    VALUE2BDD(var, var_bdd);                                                           \
    return bdd2value(bdd_##APPLY(*left_bdd, *right_bdd, value2operator(op), *var_bdd)); \
}                                                                                       \
static VALUE cBDD_##APPLY##_self(VALUE left, VALUE right, VALUE op, VALUE var)          \
{                                                                                       \
    BDD *left_bdd, *right_bdd, *var_bdd;                                                \
    VALUE2BDD(left, left_bdd);                                                           \
    VALUE2BDD(right, right_bdd);                                                          \
    VALUE2BDD(var, var_bdd);                                                               \
    assign_bdd(left_bdd,                                                                    \
	       bdd2value(bdd_##APPLY(*left_bdd, *right_bdd, value2operator(op), *var_bdd))); \
    return left;                                                                             \
}

    DEF_APPLY_OPERATOR_QUANTIFIER(appall)
    DEF_APPLY_OPERATOR_QUANTIFIER(appex)
    DEF_APPLY_OPERATOR_QUANTIFIER(appuni)


// Class methods of BDD::Pair.

static VALUE cPair_new(int argc, VALUE *argv, VALUE classValue)
{
    bddPair *bddpair = new_bddPair();
    VALUE instance = Data_Wrap_Struct(classValue, NULL, delete_bddpair, bddpair);
    rb_obj_call_init(instance, argc, argv);
    return instance;
}

// Add a pair [ old index|bdd var, new index | bdd var] to a bddPair.

static void add_pair(bddPair *bddpair, VALUE arg)
{
    VALUE old_val = rb_ary_entry(arg, 0);
    VALUE new_val = rb_ary_entry(arg, 1);
    int result;

    // Check size (we want a pair).
    if (NUM2INT(rb_funcall(arg, SIZE_ID, 0)) != 2)
    {
	rb_raise(rb_eArgError, "Pair expected.");
    }

    // What to do depends on new_val.
    if (rb_obj_is_kind_of(new_val, cBDD))
    {
	DEF_BDD(new_val, bdd);
	result = bdd_setbddpair(bddpair, check_id(old_val), *bdd);
    }
    else
    {
	result = bdd_setpair(bddpair, check_id(old_val), check_id(new_val));
    }

    if (result < 0)
    {
	rb_raise(rb_eStandardError, "Could not add the pair.");
    }
}

// Instance methods of BDD::Pair.

static VALUE cPair_initialize(int argc, VALUE *argv, VALUE self)
{
    int i;
    DEF_BDDPAIR(self, bddpair);

    // Arguments:
    // - nothing => default empty set of pairs.
    // - an array of pairs.
    // - one pair.
    // - several pairs.
    // - several arrays of pairs.
    // - 2 FDDs.

    if (argc == 2 &&
	rb_obj_is_kind_of(argv[0], cFDD) &&
	rb_obj_is_kind_of(argv[1], cFDD))
    {
	DEF_FDD(argv[0], f1);
	DEF_FDD(argv[1], f2);
	if (fdd_setpair(bddpair, f1.fdd, f2.fdd))
	{
	    rb_raise(rb_eArgError, "Invalid pair of FDDs.");
	}
	return self;
    }

    for(i = 0; i < argc; ++i)
    {
	VALUE arg = argv[i];
	if (!rb_obj_is_kind_of(arg, rb_cArray))
	{
	    rb_raise(rb_eArgError, "Array expected!");
	}
	if (rb_obj_is_kind_of(rb_ary_entry(arg, 0), rb_cArray))
	{
	    // then assume array of pairs
	    int j, size = NUM2INT(rb_funcall(arg, SIZE_ID, 0));
	    for(j = 0; j < size; ++j)
	    {
		add_pair(bddpair, rb_ary_entry(arg, j));
	    }
	}
	else // assume only one pair
	{
	    add_pair(bddpair, arg);
	}
    }

    return self;
}

static VALUE cPair_add(VALUE self, VALUE arg)
{
    DEF_BDDPAIR(self, bddpair);
    add_pair(bddpair, arg);
    return self;
}

static VALUE cPair_each(VALUE self)
{
    int i, nb = bdd_varnum();
    DEF_BDDPAIR(self, bddpair);
    assert(bddpair->result);
    for(i = 0; i < nb; ++i)
    {
	int level = bdd_var2level(i);
	if (bddpair->result[level] != bdd_ithvar(i))
	{
	    VALUE arr = rb_ary_new2(2);
	    rb_ary_store(arr, 0, INT2NUM(i));
	    rb_ary_store(arr, 1, bdd2value(bddpair->result[level]));
	    rb_yield(arr);
	}
    }
    return self;
}

static VALUE cPair_reset_self(VALUE self)
{
    DEF_BDDPAIR(self, bddpair);
    bdd_resetpair(bddpair);
    return self;
}

// Instance method of BDD::Operator.

static VALUE cOperator_to_s(VALUE self)
{
    return rb_str_new2(bdd_OP[value2operator(self)]);
}

// Buddy hooks.

static void print_handler(FILE *out, int var)
{
    VALUE val = rb_funcall(cBDD, LABEL_ID, 1, INT2NUM(var));
    char *s = StringValueCStr(val);
    if (s)
    {
	fputs(s, out);
    }
    else
    {
	fprintf(out, ":%d", var);
    }
}

static void error_handler(int error)
{
    const char *strerror = bdd_errstring(error);
    bdd_clear_error();
    rb_raise(rb_eStandardError, strerror ? strerror : "Unknown Buddy error.");
}

typedef void (*gc_handler)(int, bddGbcStat*);
gc_handler old_gbc_handler = NULL;
static int show_gc = 0;

static void gbc_handler(int pre, bddGbcStat *stat)
{
    // Call Ruby GC before Buddy's GC otherwise it's useless for Buddy.
    if (pre)
    {
	rb_funcall(rb_mGC, START_ID, 0);
    }
    // If show flag is on then call default handler (show stats).
    if (show_gc && old_gbc_handler)
    {
	old_gbc_handler(pre, stat);
    }
}

static VALUE cBDD_show_gc(VALUE self, VALUE arg)
{
    show_gc = arg == Qtrue ? 1 : 0;
    return self;
}

// BVEC

static VALUE cBVec_constant(VALUE myClass, VALUE bitVal, VALUE val)
{
    return bvec2value(bvec_con(NUM2INT(bitVal), NUM2INT(val)));
}

static VALUE cBVec_false(VALUE myClass, VALUE arg)
{
    return bvec2value(bvec_false(NUM2INT(arg)));
}

static VALUE cBVec_true(VALUE myClass, VALUE arg)
{
    return bvec2value(bvec_true(NUM2INT(arg)));
}

static VALUE cBVec_var(VALUE myClass, VALUE numVal, VALUE offVal, VALUE stepVal)
{
    int bitnum = NUM2INT(numVal);
    int offset = check_id(offVal);
    int step = NUM2INT(stepVal);
    if (offset + (bitnum-1)*step >= bdd_varnum())
    {
	rb_raise(rb_eArgError,
		 "Bit vector with undefined BDD variables.");
    }
    return bvec2value(bvec_var(bitnum, offset, step));
}

static VALUE cBVec_add(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    return bvec2value(bvec_add(*l, *r));
}

static VALUE cBVec_sub(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    return bvec2value(bvec_sub(*l, *r));
}

static VALUE cBVec_mul(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	return bvec2value(bvec_mul(*l, *r));
    }
    else
    {
	return bvec2value(bvec_mulfixed(*l, NUM2INT(arg)));
    }
}

static VALUE cBVec_shlF(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	return bvec2value(bvec_shl(*l, *r, bddfalse));
    }
    else
    {
	return bvec2value(bvec_shlfixed(*l, NUM2INT(arg), bddfalse));
    }
}

static VALUE cBVec_shl(VALUE self, VALUE arg, VALUE cVal)
{
    DEF_BVEC(self, l);
    DEF_BDD(cVal, c);
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	return bvec2value(bvec_shl(*l, *r, *c));
    }
    else
    {
	return bvec2value(bvec_shlfixed(*l, NUM2INT(arg), *c));
    }
}

static VALUE cBVec_shrF(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	return bvec2value(bvec_shr(*l, *r, bddfalse));
    }
    else
    {
	return bvec2value(bvec_shrfixed(*l, NUM2INT(arg), bddfalse));
    }
}

static VALUE cBVec_shr(VALUE self, VALUE arg, VALUE cVal)
{
    DEF_BVEC(self, l);
    DEF_BDD(cVal, c);
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	return bvec2value(bvec_shr(*l, *r, *c));
    }
    else
    {
	return bvec2value(bvec_shrfixed(*l, NUM2INT(arg), *c));
    }
}

static VALUE cBVec_div(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    BVEC res = bvec_copy(*l);
    BVEC rem = bvec_copy(*l);
    int result;
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);	
	result = bvec_div(*l, *r, &res, &rem);
    }
    else
    {
	result = bvec_divfixed(*l, NUM2INT(arg), &res, &rem);
    }
    if (result)
    {
	bvec_delref(res);
	bvec_delref(rem);
	rb_raise(rb_eStandardError, "Division error.");
    }
    bvec_delref(rem);
    return bvec2value(res);
}

static VALUE cBVec_mod(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    BVEC res = bvec_copy(*l);
    BVEC rem = bvec_copy(*l);
    int result;
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	result = bvec_div(*l, *r, &res, &rem);
    }
    else
    {
	result = bvec_divfixed(*l, NUM2INT(arg), &res, &rem);
    }
    if (result)
    {
	bvec_delref(res);
	bvec_delref(rem);
	rb_raise(rb_eStandardError, "Division error.");
    }
    bvec_delref(res);
    return bvec2value(rem);
}

static VALUE cBVec_div4(VALUE self, VALUE arg, VALUE resVal, VALUE remVal)
{
    DEF_BVEC(self, l);
    DEF_BVEC(resVal, res);
    DEF_BVEC(remVal, rem);
    int result;
    if (rb_obj_is_kind_of(arg, cBVec))
    {
	DEF_BVEC(arg, r);
	result = bvec_div(*l, *r, res, rem);
    }
    else
    {
	result = bvec_divfixed(*l, NUM2INT(arg), res, rem);
    }
    if (result)
    {
	rb_raise(rb_eStandardError, "Division error.");
    }
    return self;
}

static VALUE cBVec_eq(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    check_bvecs(l, r);
    return bdd2value(bvec_equ(*l, *r));
}

static VALUE cBVec_gte(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    check_bvecs(l, r);
    return bdd2value(bvec_gte(*l, *r));
}

static VALUE cBVec_gth(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    check_bvecs(l, r);
    return bdd2value(bvec_gth(*l, *r));
}

static VALUE cBVec_lte(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    check_bvecs(l, r);
    return bdd2value(bvec_lte(*l, *r));
}

static VALUE cBVec_lth(VALUE self, VALUE arg)
{
    DEF_BVEC(self, l);
    DEF_BVEC(arg, r);
    check_bvecs(l, r);
    return bdd2value(bvec_lth(*l, *r));
}

static VALUE cBVec_coerce(VALUE self, VALUE arg)
{
    DEF_BVEC(self, v);
    return bvec2value(bvec_coerce(NUM2INT(arg), *v));
}

static VALUE cBVec_copy(VALUE self)
{
    DEF_BVEC(self, src);
    return bvec2value(bvec_copy(*src));
}

static VALUE cBVec_isconst(VALUE self)
{
    DEF_BVEC(self, v);
    return bvec_isconst(*v) ? Qtrue : Qfalse;
}

static VALUE cBVec_ite(VALUE va, VALUE vb, VALUE vc)
{
    DEF_BDD(va, a);
    DEF_BVEC(vb, b);
    DEF_BVEC(vc, c);
    return bvec2value(bvec_ite(*a, *b, *c));
}

static VALUE cBVec_each(VALUE self)
{
    DEF_BVEC(self, v);
    int i;
    for(i = 0; i < v->bitnum; ++i)
    {
	rb_yield(bdd2value(v->bitvec[i]));
    }
    return self;
}

static VALUE cBVec_toa(VALUE self)
{
    DEF_BVEC(self, v);
    VALUE arr = rb_ary_new2(v->bitnum);
    int i;
    for(i = 0; i < v->bitnum; ++i)
    {
	rb_ary_store(arr, i, bdd2value(v->bitvec[i]));
    }
    return arr;
}

static VALUE cBVec_toi(VALUE self)
{
    DEF_BVEC(self, v);
    if (!bvec_isconst(*v))
    {
	rb_raise(rb_eArgError, "BVec is not a constant.");
    }
    return INT2NUM(bvec_val(*v));
}

static VALUE cArray2BVec(VALUE self)
{
    int i, size = NUM2INT(rb_funcall(self, SIZE_ID, 0));
    BVEC *bvec = new_BVEC();
    bvec->bitvec = (BDD*) malloc(size*sizeof(BDD));
    if (!bvec->bitvec)
    {
	free(bvec);
	rb_raise(rb_eNoMemError, "BVec not allocated!");
    }
    bvec->bitnum = size;
    for(i = 0; i < size; ++i)
    {
	DEF_BDD(rb_ary_entry(self, i), bddi);
	bdd_addref(bvec->bitvec[i] = *bddi);
    }
    return Data_Wrap_Struct(cBVec, NULL, delete_bvec, bvec);
}

// FDD.

static VALUE cFDD_size(VALUE myClass)
{
    return INT2NUM(fdd_domainnum());
}

static VALUE cFDD_add(VALUE myClass, VALUE arg)
{
    if (!rb_obj_is_kind_of(arg, rb_cArray))
    {
	int dom = NUM2INT(arg);
	if (fdd_extdomain(&dom, 1) < 0)
	{
	    rb_raise(rb_eStandardError,
		     "Finite domain block not added.");
	}
    }
    else
    {
	int i, size = NUM2INT(rb_funcall(arg, SIZE_ID, 0));
	int dom[size];
	for(i = 0; i < size; ++i)
	{
	    dom[i] = NUM2INT(rb_ary_entry(arg, i));
	}
	if (fdd_extdomain(dom, size) < 0)
	{
	    rb_raise(rb_eStandardError,
		     "Finite domain block not added.");
	}
    }
    return myClass;
}

static int check_set(VALUE arg)
{
    int var = NUM2INT(arg);
    if (var < 0 || var > fdd_domainnum())
    {
	rb_raise(rb_eArgError,
		 "Undefined domain block (%d).",
		 var);
    }
    return var;
}

static int check_var(VALUE arg)
{
    // Unknown how to check FDD variable.
    return NUM2INT(arg);
}

static VALUE cFDD_ithbddset(VALUE myClass, VALUE arg)
{
    return bdd2value(fdd_ithset(check_set(arg)));
}

static VALUE cFDD_eachbddset(VALUE myClass)
{
    int i, n = fdd_domainnum();
    for(i = 0; i < n; ++i)
    {
	rb_yield(bdd2value(fdd_ithset(i)));
    }
    return myClass;
}

static VALUE cFDD_ithset(VALUE myClass, VALUE arg)
{
    return FDD2VALUE(check_set(arg));
}

static VALUE cFDD_each(VALUE myClass)
{
    int i, n = fdd_domainnum();
    for(i = 0; i < n; ++i)
    {
	rb_yield(FDD2VALUE(i));
    }
    return myClass;
}

static VALUE cFDD_ithvar(VALUE myClass, VALUE var, VALUE val)
{
    return bdd2value(fdd_ithvar(check_var(var), NUM2INT(val)));
}

static VALUE cFDD_makeset(VALUE myClass, VALUE arg)
{
    if (!rb_obj_is_kind_of(arg, rb_cArray))
    {
	int var = check_set(arg);
	return bdd2value(fdd_makeset(&var, 1));
    }
    else
    {
	int i, size = NUM2INT(rb_funcall(arg, SIZE_ID, 0));
	int varset[size];
	for(i = 0; i < size; ++i)
	{
	    varset[i] = check_set(rb_ary_entry(arg, i));
	}
	return bdd2value(fdd_makeset(varset, size));
    }
}

/*
static VALUE cBDD_scanallvar(VALUE self)
{
    DEF_BDD(self, bdd);
    int *val = fdd_scanallvar(*bdd);
    int i, n = fdd_domainnum(); // This is wrong.
    VALUE ret = rb_ary_new2(n);
    if (val)
    {
	for(i = 0; i < n; ++i)
	{
	    rb_ary_store(ret, i, INT2NUM(val[i]));
	}
	free(val);
    }
    return ret;
}
*/

static VALUE cBDD_scanset(VALUE self)
{
    DEF_BDD(self, bdd);
    int *val;
    int i, n;
    if (!fdd_scanset(*bdd, &val, &n))
    {
	rb_raise(rb_eStandardError, "Scan set error.");
    }
    VALUE ret = rb_ary_new2(n);
    for(i = 0; i < n; ++i)
    {
	rb_ary_store(ret, i, INT2NUM(val[i]));
    }
    free(val);
    return ret;
}

static VALUE cBDD_scanvar(VALUE self, VALUE arg)
{
    DEF_BDD(self, bdd);
    int res = fdd_scanvar(*bdd, check_var(arg));
    return res < 0 ? Qnil : INT2NUM(res);
}

static VALUE cFDD_tobvec(VALUE self)
{
    DEF_FDD(self, var);
    return bvec2value(bvec_varfdd(var.fdd));
}

static VALUE cFDD_tobdd(VALUE self)
{
    DEF_FDD(self, var);
    return bdd2value(fdd_domain(var.fdd));
}

static VALUE cFDD_toi(VALUE self)
{
    DEF_FDD(self, var);
    return INT2NUM(var.fdd);
}

static VALUE cFDD_domsize(VALUE self)
{
    DEF_FDD(self, var);
    return INT2NUM(fdd_domainsize(var.fdd));
}

static VALUE cFDD_varnum(VALUE self)
{
    DEF_FDD(self, var);
    return INT2NUM(fdd_varnum(var.fdd));
}

static VALUE cFDD_equals(VALUE self, VALUE arg)
{
    DEF_FDD(self, a);
    DEF_FDD(arg, b);
    return bdd2value(fdd_equals(a.fdd, b.fdd));
}

static VALUE cFDD_overlap(VALUE self, VALUE arg)
{
    DEF_FDD(self, v1);
    DEF_FDD(arg, v2);
    return FDD2VALUE(fdd_overlapdomain(v1.fdd, v2.fdd));
}

static VALUE cFDD_tos(VALUE self)
{
    DEF_FDD(self, var);
    char buffer[32];
    snprintf(buffer, sizeof(buffer), "#%d", var.fdd);
    return rb_str_new2(buffer);
}

// Macros for declaring.

#define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)
#define METHOD(CLASS, NAME, FUNC, ARGC) rb_define_method(CLASS, NAME, VALUEFUNC(FUNC), ARGC)
#define SINGLETON(CLASS, NAME, FUNC, ARGC) rb_define_singleton_method(CLASS, NAME, VALUEFUNC(FUNC), ARGC)
#define DEF_BDD_OP(I, OP) rb_define_const(cBDD, bdd_OP[I], OPERATOR2VALUE(OP));

DLLEXPORT void Init_bdd_lib()
{
    // Start Buddy & wrapper.

    bdd_init(100000, 10000);
    SIZE_ID = rb_intern("size");
    SYSTEM_ID = rb_intern("system");
    LABEL_ID = rb_intern("label");
    START_ID = rb_intern("start");
    bdd_setcacheratio(64);
    bdd_setmaxincrease(100000);
    bdd_file_hook(print_handler);
    bdd_error_hook(error_handler);
    old_gbc_handler = bdd_gbc_hook(gbc_handler);

    // Define the class BDD.

    cBDD = rb_define_class("BDD", rb_cObject);

    SINGLETON(cBDD, "new",         cBDD_new, -1);
    SINGLETON(cBDD, "print_stat",  cBDD_print_stat, 0);
    SINGLETON(cBDD, "true",        cBDD_true, 0);
    SINGLETON(cBDD, "false",       cBDD_false, 0);
    SINGLETON(cBDD, "set_var_num", cBDD_setvarnum, 1);
    SINGLETON(cBDD, "get_var_num", cBDD_getvarnum, 0);
    SINGLETON(cBDD, "get_var",     cBDD_get_var, 1);
    SINGLETON(cBDD, "get_nvar",    cBDD_get_nvar, 1);
    SINGLETON(cBDD, "new_var",     cBDD_new_var, 0);
    SINGLETON(cBDD, "make_set",    cBDD_make_set, 1);
    SINGLETON(cBDD, "load",        cBDD_load, 1);
    SINGLETON(cBDD, "swap",        cBDD_swapvar, 2);
    SINGLETON(cBDD, "nxor",        cBDD_nxor, -1);
    SINGLETON(cBDD, "print_false", cBDD_print_false, 1);
    SINGLETON(cBDD, "show_gc",     cBDD_show_gc, 1);
    
    SINGLETON(cBDD, "set_var_id_order",    cBDD_setvarorder, 1);
    SINGLETON(cBDD, "reorder_win2",        cBDD_reorder_WIN2, 0);
    SINGLETON(cBDD, "autoreorder_win2",    cBDD_autoreorder_WIN2, 0);
    SINGLETON(cBDD, "reorder_win2ite",     cBDD_reorder_WIN2ITE, 0);
    SINGLETON(cBDD, "autoreorder_win2ite", cBDD_autoreorder_WIN2ITE, 0);
    SINGLETON(cBDD, "reorder_win3",        cBDD_reorder_WIN3, 0);
    SINGLETON(cBDD, "autoreorder_win3",    cBDD_autoreorder_WIN3, 0);
    SINGLETON(cBDD, "reorder_win3ite",     cBDD_reorder_WIN3ITE, 0);
    SINGLETON(cBDD, "autoreorder_win3ite", cBDD_autoreorder_WIN3ITE, 0);
    SINGLETON(cBDD, "reorder_sift",        cBDD_reorder_SIFT, 0);
    SINGLETON(cBDD, "autoreorder_sift",    cBDD_autoreorder_SIFT, 0);
    SINGLETON(cBDD, "reorder_siftite",     cBDD_reorder_SIFTITE, 0);
    SINGLETON(cBDD, "autoreorder_siftite", cBDD_autoreorder_SIFTITE, 0);
    SINGLETON(cBDD, "reorder_random",      cBDD_reorder_RANDOM, 0);
    SINGLETON(cBDD, "autoreorder_random",  cBDD_autoreorder_RANDOM, 0);
    SINGLETON(cBDD, "stop_autoreorder",    cBDD_autoreorder_NONE, 0);

    METHOD(cBDD, "initialize", cBDD_initialize, -1);
//  METHOD(cBDD, "print_all",  cBDD_print_all, 0);
//  METHOD(cBDD, "print_one",  cBDD_print_one, 0);
    METHOD(cBDD, "node_count", cBDD_node_count, 0);
    METHOD(cBDD, "path_count", cBDD_path_count, 0);
    METHOD(cBDD, "id",         cBDD_id, 0);
    METHOD(cBDD, "==",         cBDD_equal, 1);
    METHOD(cBDD, "true?",      cBDD_is_true, 0);
    METHOD(cBDD, "false?",     cBDD_is_false, 0);
    METHOD(cBDD, "var?",       cBDD_is_var, 0);
    METHOD(cBDD, "nvar?",      cBDD_is_nvar, 0);
    METHOD(cBDD, "var_id",     cBDD_var, 0);
    METHOD(cBDD, "low",        cBDD_low, 0);
    METHOD(cBDD, "high",       cBDD_high, 0);
    METHOD(cBDD, "to_ps",      cBDD_to_ps, 1);
    METHOD(cBDD, "save",       cBDD_save, 1);
    METHOD(cBDD, "load!",      cBDD_load_self, 1);
    METHOD(cBDD, "support",    cBDD_support, 0);

    METHOD(cBDD, "simplify",     cBDD_simplify, 1);
    METHOD(cBDD, "simplify!",    cBDD_simplify_self, 1);
    METHOD(cBDD, "restrict",     cBDD_restrict, 1);
    METHOD(cBDD, "restrict!",    cBDD_restrict_self, 1);
    METHOD(cBDD, "exist",        cBDD_exist, 1);
    METHOD(cBDD, "exist!",       cBDD_exist_self, 1);
    METHOD(cBDD, "forall",       cBDD_forall, 1);
    METHOD(cBDD, "forall!",      cBDD_forall_self, 1);
    METHOD(cBDD, "unique",       cBDD_unique, 1);
    METHOD(cBDD, "unique!",      cBDD_unique_self, 1);
    METHOD(cBDD, "replace",      cBDD_replace, 1);
    METHOD(cBDD, "replace!",     cBDD_replace_self, 1);
    METHOD(cBDD, "compose",      cBDD_compose, 2);
    METHOD(cBDD, "compose!",     cBDD_compose_self, 2);
    METHOD(cBDD, "sim_compose",  cBDD_sim_compose, 2);
    METHOD(cBDD, "sim_compose!", cBDD_sim_compose_self, 2);
    METHOD(cBDD, "ite",          cBDD_ite, 2);
    METHOD(cBDD, "ite!",         cBDD_ite_self, 2);
    METHOD(cBDD, "constrain",    cBDD_constrain, 1);
    METHOD(cBDD, "constrain!",   cBDD_constrain_self, 1);

    METHOD(cBDD, "each",         cBDD_each, 0);
    METHOD(cBDD, "sat_count",    cBDD_satcount, 0);
    METHOD(cBDD, "sat_countln",  cBDD_satcountln, 0);
    METHOD(cBDD, "sat_one",      cBDD_satone, 0);
    METHOD(cBDD, "full_sat_one", cBDD_fullsatone, 0);

    METHOD(cBDD, "not",     cBDD_not, 0);
    METHOD(cBDD, "not!",    cBDD_not_self, 0);
    METHOD(cBDD, "&",       cBDD_and, 1);
    METHOD(cBDD, "and",     cBDD_and, 1);
    METHOD(cBDD, "and!",    cBDD_and_self, 1);
    METHOD(cBDD, "^",       cBDD_xor, 1);
    METHOD(cBDD, "xor",     cBDD_xor, 1);
    METHOD(cBDD, "xor!",    cBDD_xor_self, 1);
    METHOD(cBDD, "|",       cBDD_or, 1);
    METHOD(cBDD, "or",      cBDD_or, 1);
    METHOD(cBDD, "or!",     cBDD_or_self, 1);
    METHOD(cBDD, "nand",    cBDD_nand, 1);
    METHOD(cBDD, "nand!",   cBDD_nand_self, 1);
    METHOD(cBDD, "nor",     cBDD_nor, 1);
    METHOD(cBDD, "nor!",    cBDD_nor_self, 1);
    METHOD(cBDD, ">>",      cBDD_imp, 1);
    METHOD(cBDD, "imply",   cBDD_imp, 1);
    METHOD(cBDD, "imply!",  cBDD_imp_self, 1);
    METHOD(cBDD, "<=>",     cBDD_biimp, 1);
    METHOD(cBDD, "biimp",   cBDD_biimp, 1);
    METHOD(cBDD, "biimp!",  cBDD_biimp_self, 1);
    METHOD(cBDD, "-",       cBDD_diff, 1);
    METHOD(cBDD, "diff",    cBDD_diff, 1);
    METHOD(cBDD, "diff!",   cBDD_diff_self, 1);
    METHOD(cBDD, "<",       cBDD_less, 1);
    METHOD(cBDD, "less",    cBDD_less, 1);
    METHOD(cBDD, "less!",   cBDD_less_self, 1);
    METHOD(cBDD, "<<",      cBDD_invimp, 1);
    METHOD(cBDD, "invimp",  cBDD_invimp, 1);
    METHOD(cBDD, "invimp!", cBDD_invimp_self, 1);

    METHOD(cBDD, "apply",         cBDD_apply, 2);
    METHOD(cBDD, "apply!",        cBDD_apply_self, 2);
    METHOD(cBDD, "apply_forall",  cBDD_appall, 3);
    METHOD(cBDD, "apply_forall!", cBDD_appall_self, 3);
    METHOD(cBDD, "apply_exist",   cBDD_appex, 3);
    METHOD(cBDD, "apply_exist!",  cBDD_appex_self, 3);
    METHOD(cBDD, "apply_unique",  cBDD_appuni, 3);
    METHOD(cBDD, "apply_unique!", cBDD_appuni_self, 3);
    //METHOD(cBDD, "scanallvar",    cBDD_scanallvar, 0);
    METHOD(cBDD, "scanset",       cBDD_scanset, 0);
    METHOD(cBDD, "scanvar",       cBDD_scanvar, 1);

    // Define the class BDD::Pair.

    cPair = rb_define_class_under(cBDD, "Pair", rb_cObject);
    SINGLETON(cPair, "new", cPair_new, -1);

    METHOD(cPair, "initialize", cPair_initialize, -1);    
    METHOD(cPair, "<<",         cPair_add, 1);
    METHOD(cPair, "each",       cPair_each, 0);
    METHOD(cPair, "reset!",     cPair_reset_self, 0);

    // Define the class BDD::Operation and its constant values.
    // There is no new because the user is not supposed to create them.

    cOperator = rb_define_class_under(cBDD, "Operator", rb_cObject);
    METHOD(cOperator, "to_s", cOperator_to_s, 0);
    DEF_BDD_OP(0, bddop_and);
    DEF_BDD_OP(1, bddop_xor);
    DEF_BDD_OP(2, bddop_or);
    DEF_BDD_OP(3, bddop_nand);
    DEF_BDD_OP(4, bddop_nor);
    DEF_BDD_OP(5, bddop_imp);
    DEF_BDD_OP(6, bddop_biimp);
    DEF_BDD_OP(7, bddop_diff);
    DEF_BDD_OP(8, bddop_less);
    DEF_BDD_OP(9, bddop_invimp);

    // Define the class BVEC.

    cBVec = rb_define_class("BVec", rb_cObject);
    SINGLETON(cBVec, "constant", cBVec_constant, 2);
    SINGLETON(cBVec, "false",    cBVec_false, 1);
    SINGLETON(cBVec, "true",     cBVec_true, 1);
    SINGLETON(cBVec, "var",      cBVec_var, 3);

    METHOD(cBVec, "+",         cBVec_add, 1);
    METHOD(cBVec, "-",         cBVec_sub, 1);
    METHOD(cBVec, "*",         cBVec_mul, 1);
    METHOD(cBVec, "/",         cBVec_div, 1);
    METHOD(cBVec, "%",         cBVec_mod, 1);
    METHOD(cBVec, "==",        cBVec_eq, 1);
    METHOD(cBVec, ">=",        cBVec_gte, 1);
    METHOD(cBVec, ">",         cBVec_gth, 1);
    METHOD(cBVec, "<=",        cBVec_lte, 1);
    METHOD(cBVec, "<",         cBVec_lth, 1);
    METHOD(cBVec, "<<",        cBVec_shlF, 1);
    METHOD(cBVec, ">>",        cBVec_shrF, 1);
    METHOD(cBVec, "shl",       cBVec_shl, 2);
    METHOD(cBVec, "shr",       cBVec_shr, 2);
    METHOD(cBVec, "div",       cBVec_div4, 3);
    METHOD(cBVec, "coerce",    cBVec_coerce, 1);
    METHOD(cBVec, "copy",      cBVec_copy, 0);
    METHOD(cBVec, "constant?", cBVec_isconst, 0);
    METHOD(cBVec, "ite",       cBVec_ite, 2);
    METHOD(cBVec, "each",      cBVec_each, 0);
    METHOD(cBVec, "to_a",      cBVec_toa, 0);
    METHOD(cBVec, "to_i",      cBVec_toi, 0);

    // Conversion Array -> BVec.
    METHOD(rb_cArray, "to_bvec", cArray2BVec, 0);

    // Define the class FDD.

    cFDD = rb_define_class("FDD", rb_cObject);
    SINGLETON(cFDD, "size",     cFDD_domsize, 0);
    SINGLETON(cFDD, "add",      cFDD_add, 1);
    SINGLETON(cFDD, "[]",       cFDD_ithset, 1);
    SINGLETON(cFDD, "each",     cFDD_each, 0);
    SINGLETON(cFDD, "set",      cFDD_ithbddset, 1);
    SINGLETON(cFDD, "each_set", cFDD_eachbddset, 0);
    SINGLETON(cFDD, "var",      cFDD_ithvar, 2);
    SINGLETON(cFDD, "make_set", cFDD_makeset, 1);

    METHOD(cFDD, "to_bvec", cFDD_tobvec, 0); // FDD -> BVec.
    METHOD(cFDD, "to_bdd",  cFDD_tobdd, 0); // FDD -> BDD.
    METHOD(cFDD, "to_i",    cFDD_toi, 0);
    METHOD(cFDD, "size",    cFDD_size, 0);
    METHOD(cFDD, "var_num", cFDD_varnum, 0);
    METHOD(cFDD, "equals",  cFDD_equals, 1);
    METHOD(cFDD, "+",       cFDD_overlap, 1);
    METHOD(cFDD, "to_s",    cFDD_tos, 0);
}
