# Load Buddy wrapper.
require 'bdd_lib'

# Add support for named variables. Names are referred
# as keys since they are not necessarily strings.
class BDD
  include Enumerable

  @@variables = Hash.new # Table of variable IDs (key => ID).
  @@keys = []            # Table of variable keys (ID => key).

  # Get/create a variable ID.
  def BDD.var_id(key)
    return @@variables[key] if @@variables[key]
    @@keys << key
    BDD.set_var_num(@@keys.size)
    @@variables[key] = @@variables.size
  end

  # Get/create a variable with a given key.
  def BDD.var(key)
    BDD.get_var(BDD.var_id(key))
  end

  # Get/create a negated variable with a given key.
  def BDD.nvar(key)
    BDD.get_nvar(BDD.var_id(key))
  end

  # Get the key of a variable ID.
  def BDD.key(id)
    @@keys[id]
  end

  # Get the label of a variable ID.
  def BDD.label(id)
    (id >= 0 && id < @@keys.size) ? @@keys[id].to_s : ":" + id.to_s
  end

  # Change variable ordering (access by key).
  def BDD.set_var_key_order(keys)
    BDD.set_var_id_order(keys.collect { |k| BDD.var_id(k) })
  end

  # Change variable ordering (access by variable = BDD).
  def BDD.set_var_order(vars)
    BDD.set_var_id_order(vars.collect { |v| v.var_id })
  end

  # Convert BDDs to strings.
  def dump
    return "true" if true?
    return "false" if false?
    "(" + label + " & " + high.dump + ") | (!" + label + " & " + low.dump + ")"
  end
  def to_s
    return "true" if true?
    return "false" if false?
    return "..." if big?
    if (high <=> BDD.true).true?
      h = label
    elsif (high <=> BDD.false).true?
      h = nil
    else
      h = "(" + label + " & " + high.to_s + ")"
    end
    if (low <=> BDD.true).true?
      l = "!" + label
    elsif (low <=> BDD.false).true?
      l = nil
    else
      l = "(!" + label + " & " + low.to_s + ")"
    end
    return h if !l
    return l if !h
    return h + " | " + l
  end

  # Subjective judgement - you can overwrite this to tune to_s.
  def big?
    node_count > 100
  end

  # This is to print BDDs in irb.
  def inspect
    "#<#{self.class.to_s}:#{id} \"" + (node_count < 100 ? "#{to_s}\">" : "...\">")
  end

  # Show this BDD with a postscript viewer.
  def show(*a)
    filename="/tmp/ruby_bdd_output" + rand.to_s
    to_ps(filename)
    Thread.new(a.size > 0 ? a[0] : "gv") { |viewer|
      system(viewer + " " + filename + " 1>/dev/null && rm -f " + filename)
    }
  end

  # Label of this BDD. The terminals true and false have no label.
  def label
    BDD.label(var_id)
  end

  # Add some convertions.
  class Pair
    include Enumerable

    def to_a
      collect { |id_old, bdd_new| [ id_old, bdd_new ] }
    end

    def to_s
      "{" + collect { |id_old, bdd_new| "(#{BDD.label(id_old)},#{bdd_new.to_s})" }.join(",") + "}"
    end

    def inspect
      "#<#{self.class.to_s} #{to_s}>"
    end
  end

  # Meaningful inspect in irb.
  class Operator
    def inspect
      "#<#{self.class.to_s} #{to_s}>"
    end
  end
end

class BVec
  include Enumerable
end

# Add support for conversion to BDD::Pair and to
# arrays of key strings.
class Array
  # Conversion of (array of) [id|bdd var, bdd] to pairs.
  def to_pair
    BDD::Pair.new(self)
  end

  # Conversion to a set of variables.
  def to_set
    BDD.make_set(self)
  end

  # Conversion of truth array to variable keys.
  def to_keys
    k = i = 0
    collect { |b|
      k = BDD.key(i)
      i += 1
      k && b != nil ? (b ? k.to_s : "!#{k.to_s}") : b
    }
  end
end

# Shortcuts.

def var(key)
  BDD.var(key)
end

def nvar(key)
  BDD.nvar(key)
end

def vars(*v)
  v.collect { |i| "#{i}=BDD.var('#{i}')" }.join("\n")
end
